"""SOCKS proxy server implementation with multi-process support.

This module implements a multi-process SOCKS proxy server with the following features:
- Process pool management for handling multiple connections
- Automatic process recovery on failure
- WiFi interface detection and validation
- Clean shutdown handling
- UI integration
- Clipboard integration for easy sharing

The server supports:
- Multiple simultaneous connections
- Process-level parallelism
- Automatic port reuse
- Privilege checking
- Interface validation

Example:
    # Create and start a proxy server with 4 processes
    create_proxy_server("192.168.1.100", 9050, 4)
"""

import contextlib
import multiprocessing
import os
import socket
import socketserver
import time
from typing import Final

import pyperclip
from loguru import logger
from rich.console import Console

from hotspot_socks_proxy.core.utils.prompt.proxy_ui import create_proxy_ui
from hotspot_socks_proxy.core.network import get_active_interfaces
from hotspot_socks_proxy.core.utils.routing import setup_routing, cleanup_routing

from .socks_handler import SocksHandler

console = Console()

# Type aliases
ProcessList = list[multiprocessing.Process]

# Constants
BANDWIDTH_THRESHOLD: Final = 100  # Bytes
PROCESS_JOIN_TIMEOUT: Final = 1.0  # Seconds
STARTUP_DELAY: Final = 0.1  # Seconds between process starts
CLIPBOARD_DELAY: Final = 0.5  # Seconds to wait after clipboard copy


class SocksProxy(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """SOCKS proxy server implementation."""

    allow_reuse_address = True
    daemon_threads = True
    request_queue_size = 100

    def __init__(self, server_address: tuple[str, int], handler_class, interface_name: str) -> None:
        """Initialize the SOCKS proxy server.
        
        Args:
            server_address: Tuple of (host, port) to bind to
            handler_class: Handler class for processing requests
            interface_name: Name of the network interface to use
        """
        self.interface_name = interface_name  # Store the interface name
        super().__init__(server_address, handler_class)

    def server_bind(self) -> None:
        """Bind the server socket with reuse options."""
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, "SO_REUSEPORT"):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        super().server_bind()


def run_server(host: str, port: int, interface_name: str) -> None:
    """Individual process server runner.

    Args:
        host: Host address to bind to
        port: Port number to listen on
        interface_name: Name of the network interface to use
    """
    server: SocksProxy | None = None
    try:
        if os.name != "nt" and os.geteuid() != 0:
            logger.warning("Running without root privileges")
            console.print(
                "[yellow]Warning: Running without root privileges may limit functionality."
            )

        server = SocksProxy((host, port), SocksHandler, interface_name)
        logger.info(f"Server process started on {host}:{port} using interface {interface_name}")
        server.serve_forever()
    except KeyboardInterrupt:
        logger.info("Server process stopping")
    except Exception as e:
        logger.exception("Server process error")
        console.print(f"[red]Server error: {e}")
    finally:
        if server:
            with contextlib.suppress(Exception):
                server.server_close()
                logger.info("Server closed")


def create_proxy_server(host: str, port: int, num_processes: int) -> None:
    """Create and start a multi-process SOCKS proxy server."""
    processes: ProcessList = []

    try:
        # Get interface name from network module
        interface = next(
            (iface for iface in get_active_interfaces() if iface.ip == host),
            None
        )
        if not interface:
            raise RuntimeError(f"No interface found with IP {host}")

        # Initialize UI
        ui_thread = create_proxy_ui(host, port)
        if ui_thread:
            ui_thread.start()

        # Set up routing rules for the interface
        if not setup_routing(interface.name, host):
            raise RuntimeError("Failed to set up routing rules")

        # Copy address to clipboard
        try:
            proxy_address = f"socks5://{host}:{port}"  # Added SOCKS5 protocol
            pyperclip.copy(proxy_address)
            console.print("[bold green]SOCKS5 proxy address copied to clipboard")
            time.sleep(CLIPBOARD_DELAY)
        except Exception as e:
            console.print(f"[yellow]Could not copy to clipboard: {e}")

        # Start worker processes
        console.print(f"[bold green]Starting {num_processes} worker processes...")
        for _ in range(num_processes):
            process = multiprocessing.Process(
                target=run_server,
                args=(host, port, interface.name)
            )
            process.daemon = True
            processes.append(process)
            process.start()
            time.sleep(STARTUP_DELAY)

        console.print(f"\n[bold green]SOCKS5 proxy server running on {host}:{port}")
        console.print("[yellow]Use Ctrl+C to stop the server\n")

        # Keep the main process running until interrupted
        try:
            while True:
                # Check if any process died and restart it
                for i, process in enumerate(processes):
                    if not process.is_alive():
                        logger.warning(f"Process {i} died, restarting...")
                        process.join(timeout=PROCESS_JOIN_TIMEOUT)
                        new_process = multiprocessing.Process(
                            target=run_server,
                            args=(host, port, interface.name)
                        )
                        new_process.daemon = True
                        new_process.start()
                        processes[i] = new_process
                time.sleep(1)  # Check every second
        except KeyboardInterrupt:
            logger.info("Received shutdown signal")
            console.print("\n[yellow]Shutting down proxy server...")

    except Exception as e:
        logger.exception("Error in proxy server")
        console.print(f"[red]Error: {e}")
    finally:
        # Clean up routing rules
        cleanup_routing(host)
        
        logger.info("Cleaning up processes...")
        for process in processes:
            if process.is_alive():
                process.terminate()
                process.join(timeout=PROCESS_JOIN_TIMEOUT)
                if process.is_alive():
                    logger.warning("Force killing process...")
                    process.kill()
        logger.info("All processes terminated")
